package com.tzld.longarticle.recommend.server.service.recall;

import com.tzld.longarticle.recommend.server.common.ThreadPoolFactory;
import com.tzld.longarticle.recommend.server.model.Content;
import com.tzld.longarticle.recommend.server.model.ContentHisPublishArticle;
import com.tzld.longarticle.recommend.server.remote.AIGCRemoteService;
import com.tzld.longarticle.recommend.server.repository.aigc.CrawlerMetaArticleRepository;
import com.tzld.longarticle.recommend.server.repository.crawler.ArticleRepository;
import com.tzld.longarticle.recommend.server.repository.entity.aigc.CrawlerMetaArticle;
import com.tzld.longarticle.recommend.server.repository.entity.crawler.Article;
import com.tzld.longarticle.recommend.server.service.recall.strategy.ColdStartBackupRecallStrategy;
import com.tzld.longarticle.recommend.server.service.recall.strategy.DefaultRecallStrategy;
import com.tzld.longarticle.recommend.server.util.CommonCollectionUtils;
import com.tzld.longarticle.recommend.server.util.JSONUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import java.util.*;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * @author dyp
 */
@Service
@Slf4j
public class RecallService implements ApplicationContextAware {

    @Autowired
    ArticleRepository articleRepository;
    @Autowired
    CrawlerMetaArticleRepository crawlerMetaArticleRepository;
    @Autowired
    private AIGCRemoteService aigcRemoteService;

    private final Map<String, RecallStrategy> strategyMap = new HashMap<>();
    private ApplicationContext applicationContext;
    private final ExecutorService pool = ThreadPoolFactory.recallPool();

    @PostConstruct
    public void init() {
        Map<String, RecallStrategy> type = applicationContext.getBeansOfType(RecallStrategy.class);
        for (Map.Entry<String, RecallStrategy> entry : type.entrySet()) {
            RecallStrategy value = entry.getValue();
            strategyMap.put(value.getClass().getSimpleName(), value);
        }
    }

    public RecallResult recall(RecallParam param) {
        log.info("RecallParam {}", JSONUtils.toJson(param));
        List<RecallStrategy> strategies = getRecallStrategy(param);
        log.info("RecallStrategy {}", JSONUtils.toJson(CommonCollectionUtils.toList(strategies,
                s -> s.getClass().getSimpleName())));
        List<Content> content = getAllContent(param);
        param.setContent(content);
        CountDownLatch cdl = new CountDownLatch(strategies.size());
        List<Future<RecallResult.RecallData>> recallResultFutures = new ArrayList<>();
        for (final RecallStrategy strategy : strategies) {
            Future<RecallResult.RecallData> future = pool.submit(() -> {
                try {
                    return strategy.recall(param);
                } finally {
                    cdl.countDown();
                }
            });
            recallResultFutures.add(future);
        }
        try {
            cdl.await(10000, TimeUnit.MILLISECONDS);
        } catch (InterruptedException e) {
            log.error("recall error", e);
            return null;
        }

        List<RecallResult.RecallData> results = new ArrayList<>();
        for (Future<RecallResult.RecallData> f : recallResultFutures) {
            try {
                RecallResult.RecallData data = f.get();
                results.add(data);
            } catch (Exception e) {
                log.error("future get error ", e);
            }
        }

        return new RecallResult(results);
    }

    private List<RecallStrategy> getRecallStrategy(RecallParam param) {
        List<RecallStrategy> strategies = new ArrayList<>();
        strategies.add(strategyMap.get(DefaultRecallStrategy.class.getSimpleName()));
        strategies.add(strategyMap.get(ColdStartBackupRecallStrategy.class.getSimpleName()));
        return strategies;
    }

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        this.applicationContext = applicationContext;
    }

    private List<Content> getAllContent(RecallParam param) {
        List<Content> content = aigcRemoteService.getAllContent(param);
        // 标题历史均值
        setTitleAvgViewCount(content);
        // category 查询
        setContentCategory(content);
        return content;
    }

    public void setContentCategory(List<Content> contentList) {
        long start = System.currentTimeMillis();
        List<String> channelContentIds = contentList.stream().map(Content::getCrawlerChannelContentId).collect(Collectors.toList());
        List<CrawlerMetaArticle> categoryList = getContentCategoryByChannelContentId(channelContentIds);
        if (CollectionUtils.isEmpty(categoryList)) {
            return;
        }
        Map<String, List<String>> categoryMap = categoryList.stream().collect(Collectors.groupingBy(CrawlerMetaArticle::getChannelContentId,
                Collectors.mapping(CrawlerMetaArticle::getCategory, Collectors.toList())));
        for (Content content : contentList) {
            content.setCategory(categoryMap.get(content.getCrawlerChannelContentId()));
        }
        log.info("setContentCategory cost:{}", System.currentTimeMillis() - start);
    }

    private List<CrawlerMetaArticle> getContentCategoryByChannelContentId(List<String> channelContentIds) {
        if (CollectionUtils.isEmpty(channelContentIds)) {
            return new ArrayList<>();
        }
        return crawlerMetaArticleRepository.getByChannelContentIdIn(channelContentIds);
    }

    public void setTitleAvgViewCount(List<Content> contentList) {
        long start = System.currentTimeMillis();
        List<String> titleList = contentList.stream().map(Content::getTitle).collect(Collectors.toList());
        List<Article> hisArticleList = articleRepository.getByTitleIn(titleList);
        Map<String, Map<Integer, List<Article>>> map = hisArticleList.stream()
                .collect(Collectors.groupingBy(Article::getTitle, Collectors.groupingBy(Article::getItemIndex)));
        for (Content content : contentList) {
            List<Article> hisArticles = new ArrayList<>();
            Map<Integer, Integer> indexAvgMap = new HashMap<>();
            Map<Integer, List<Article>> indexArticleMap = map.get(content.getTitle());
            if (Objects.nonNull(indexArticleMap) && !indexArticleMap.isEmpty()) {
                for (Map.Entry<Integer, List<Article>> indexArticleEntry : indexArticleMap.entrySet()) {
                    List<Article> indexArticleList = indexArticleEntry.getValue();
                    int sumViewCount = indexArticleList.stream().mapToInt(Article::getShowViewCount).sum();
                    indexAvgMap.put(indexArticleEntry.getKey(), sumViewCount/ indexArticleList.size());
                    hisArticles.addAll(indexArticleList);
                }
            }
            content.setHisPublishArticleList(new ArrayList<>());
            for (Article hisArticle : hisArticles) {
                ContentHisPublishArticle article = new ContentHisPublishArticle();
                BeanUtils.copyProperties(hisArticle, article);
                article.setAvgViewCount(indexAvgMap.get(article.getItemIndex()));
                if (Objects.nonNull(article.getAvgViewCount()) && article.getAvgViewCount() > 0) {
                    article.setViewCountRate((article.getShowViewCount() * 1.0) / article.getAvgViewCount());
                }
                content.getHisPublishArticleList().add(article);
            }
        }
        log.info("setTitleAvgViewCount cost:{}", System.currentTimeMillis() - start);
    }

}
